import choco.mem.*;
import choco.*;
import choco.prop.*;
import choco.util.*;
import choco.integer.*;
import choco.palm.*;
import choco.palm.explain.*;
import choco.palm.integer.*;
import choco.palm.search.*;
import org.aspectj.lang.*;

/*
 * Created on 6 oct. 2004
 */

/**
 * Oadymppac trace implementation for Palm (DTD v2.1,
 * cf http://contraintes.inria.fr/OADymPPaC/).
 * 
 * Implemented by Guillaume Rochart (EMN/LINA)
 */
public aspect PalmTrace extends AbstractTrace {
	// Previous domain
	protected int[]  previousDomain;
	
	static {
		eventConfig = new boolean[]{
				true,  // New var
				true,  // New cst
				true,  // Choice Point
				false, // Back-to
				true,  // Solution
				true,  // Failure
				true,  // Reduce
				true,  // Restore
				true,  // Explain
				true,  // Post
				true,  // Remove
				true,  // Context
				true   // Awake
		};
		
		attributeConfig = new boolean[]{
				true,  // Chrono
				false, // Depth
				false, // Time
				false, // Line
				false  // File
		};
	}
	
	// ==============================================
	// |                                            |
	// |                 PointCuts                  |
	// |                                            |
	// ==============================================
	
	// Posts during solving
	
	pointcut newIndConstraintInSolve(Constraint c, Explanation e): 
		execution(public void Problem+.post(Constraint, Explanation)) 
		&& args(c,e) && cflow(execution (* Problem+.solve*(..)));
	
	pointcut newConstraintInSolve(Constraint c): 
		execution(public void Problem+.post(Constraint)) 
		&& args(c) && cflow(execution (*  Problem+.solve*(..)))
		&& !cflow(execution(public void Problem+.post(Constraint, Explanation)));
	
	// Backward choices
	
	pointcut backChoicePoint(Constraint c, Constraint indirect): 
		execution (* Constraint+.setPassive()) && target(indirect) &&
		cflow(execution(* PalmProblem+.remove(..)) && args(c));
	
	// Any modification
	
	pointcut anyUpdate(PalmIntVar var, Explanation e): 
		(execution(* IntVar+.update*(..)) || execution(* IntVar+.remove*(..)) 
				|| execution(* IntVar+.instantiate(..))) && target(var) && args(*,*,e);
	
	// Restoration
	
	pointcut restoreInf(Constraint cst, PalmIntVar var, int newVal):
		execution(* IntVar+.restoreInf(..)) && args(newVal) && target(var) &&
		cflow(execution(* Problem+.remove(..)) && args(cst));
		
	pointcut restoreSup(Constraint cst, PalmIntVar var, int newVal):
		execution(* IntVar+.restoreSup(..)) && args(newVal) && target(var) &&
		cflow(execution(* Problem+.remove(..)) && args(cst));
	
	pointcut restoreVal(Constraint cst, PalmIntVar var, int newVal):
		execution(* IntVar+.restoreVal(..)) && args(newVal) && target(var) &&
		cflow(execution(* Problem+.remove(..)) && args(cst));
	
	// Bound updates
	
	pointcut updateInf(Constraint cst, PalmIntVar var, Explanation e): 
		execution(* ChocEngine+.postUpdateInf(..)) && args(var, *) &&
	    cflow(execution(* *.updateInf(..)) && target(PalmIntervalIntDomain+) && args(*,*,e)) &&
		cflow(anyAwake() && target(cst));

	pointcut updateSup(Constraint cst, PalmIntVar var, Explanation e): 
		execution(* ChocEngine+.postUpdateSup(..)) && args(var, *) &&
	    cflow(execution(* *.updateSup(..)) && target(PalmIntervalIntDomain+) && args(*,*,e)) &&
		cflow(anyAwake() && target(cst));
	
	// Removals
	
	pointcut removeVal(Constraint cst, PalmIntVar var, Explanation e, int val): 
		execution(* ChocEngine+.postRemoveVal(..)) && args(var, val, *) &&
	    cflow(execution(* *.removeVal(..)) && target(PalmBitSetIntDomain+) && args(*,*,e)) &&
		cflow(anyAwake() && target(cst));
	
	
	
	// ==============================================
	// |                                            |
	// |                  Advices                   |
	// |                                            |
	// ==============================================
	
	// *********** POST ***********
	
	before(Constraint c, Explanation e): newIndConstraintInSolve(c, e) {
		c.constNumber = constNb++;
		elementOpening("<new-constraint " + eventAttributes(thisJoinPointStaticPart) + 
			" cident=\"c" + (c.constNumber) +
			"\" cexternal=\"" + c +
			"\" orig=\"system\">");
		elementClosing("</new-constraint>");
		singleElement("<post " + eventAttributes(thisJoinPointStaticPart) + 
			" cident=\"c" + (c.constNumber) + "\" context=\"");
		java.util.BitSet csts = e.getBitSet();
		for (int i = csts.nextSetBit(0); i >= 0; i = csts.nextSetBit(i + 1)) {
			stream.print("c" + ((PalmProblem)c.getProblem()).getConstraintNb(i).constNumber + " ");
		}
		stream.println("\"/>");
		stream.flush();
	}
	
	before(Constraint c): newConstraintInSolve(c) {
		c.constNumber = constNb++;
		singleElementLn("<choice-point " + eventAttributes(thisJoinPointStaticPart) + "/>");
		elementOpening("<new-constraint " + eventAttributes(thisJoinPointStaticPart) +
			" cident=\"c" + (c.constNumber) +
			"\" cexternal=\"" + c +
			"\" orig=\"system\">");
		elementClosing("</new-constraint>");
		singleElementLn("<post " + eventAttributes(thisJoinPointStaticPart) +
			" cident=\"c" + (c.constNumber) +
			"\"/>");
		stream.flush();
	}
	
	// *********** REMOVE ***********
	
	before(Constraint c, Constraint indirect): backChoicePoint(c, indirect) {
		if (c==indirect) singleElementLn("<choice-point " +
				eventAttributes(thisJoinPointStaticPart) + "/>");
		singleElementLn("<remove " + eventAttributes(thisJoinPointStaticPart) +
			" cident=\"c" + (indirect.constNumber) +
			((c != indirect)?("\" context=\"c" + (c.constNumber)):"") +
			"\"/>");
	}
	
	// Domain storing before modifications...
	
	before(IntVar var, Explanation e): anyUpdate(var, e) {
		if (var.hasEnumeratedDomain()) {
			previousDomain = new int[var.getDomainSize()];
			int i = 0;
			for(IntIterator iterator = var.getDomain().getIterator(); iterator.hasNext(); ) {
				previousDomain[i++] = iterator.next();
			}
		} else {
			previousDomain = new int[]{var.getInf(), var.getSup()};
		}
	}
	
	// *********** FAILURE ***********
	
	after(IntVar var, Explanation e) throwing (PalmContradiction p) : anyUpdate(var, e) {
		elementOpening("<reduce " + eventAttributes(thisJoinPointStaticPart) + ">");
		elementOpening("<delta>");
		if (var.hasEnumeratedDomain()) {
			singleElement("<values>");
			for(int i = 0; i < previousDomain.length; i ++) {
				stream.print(previousDomain[i] + " ");
			}
			stream.println("</values>");
		} else {
			singleElementLn("<range from=\""+previousDomain[0]+"\" to=\""+previousDomain[1]+"\"/>");
		}
		elementClosing("</delta>");
		singleElementLn("<update vident=\"v" + var.varNumber + "\" types=\"empty\"/>");
		elementOpening("<explanation>");
		if (var.hasEnumeratedDomain()) {
			singleElement("<values>");
			for(int i = 0; i < previousDomain.length; i ++) {
				stream.print(previousDomain[i] + " ");
			}
			stream.println("</values>");
		} else {
			singleElementLn("<range from=\""+previousDomain[0]+"\" to=\""+previousDomain[1]+"\"/>");
		}
		singleElement("<constraints cidents=\"");
		java.util.BitSet csts = e.getBitSet();
		for (int i = csts.nextSetBit(0); i >= 0; i = csts.nextSetBit(i + 1)) {
			stream.print("c" + ((PalmProblem)var.getProblem()).getConstraintNb(i).constNumber + " ");
		}
		stream.println("\"/>");
		elementClosing("</explanation>");
		elementClosing("</reduce>");
		singleElementLn("<failure " + eventAttributes(thisJoinPointStaticPart) + "/>");
		stream.flush();
	}
	
	// *********** RESTORE ***********
	
	before(Constraint cst, PalmIntVar var, int newVal): restoreInf(cst, var, newVal) {
		restoreBoundHandle(cst, var, newVal, 0, thisJoinPointStaticPart);
	}
	
	before(Constraint cst, PalmIntVar var, int newVal): restoreSup(cst, var, newVal) {
		restoreBoundHandle(cst, var, newVal, 1, thisJoinPointStaticPart);
	}
	
	before(Constraint cst, PalmIntVar var, int newVal): restoreVal(cst, var, newVal) {
		elementOpening("<restore " + eventAttributes(thisJoinPointStaticPart) + ">");
		elementOpening("<delta vident=\"v" + var.varNumber + "\">");
		singleElementLn("<values>" + newVal + "</values>");
		elementClosing("</delta>");
		singleElementLn("<update  vident=\"v" + var.varNumber + "\" types=\"val\"/>");
		elementClosing("</restore>");
	}	
	
	// *********** REDUCE ***********
	
	before(Constraint cst, PalmIntVar var, Explanation e): updateInf(cst, var, e) {
		updateBoundHandle(cst, var, e, 0, thisJoinPointStaticPart);
	}

	before(Constraint cst, PalmIntVar var, Explanation e): updateSup(cst, var, e) {
		updateBoundHandle(cst, var, e, 1, thisJoinPointStaticPart);
	}
	
	before(Constraint cst, PalmIntVar var, Explanation e, int val): removeVal(cst, var, e, val) {
		removeValHandle(cst, var, e, val, thisJoinPointStaticPart);
	}
	
	// ==============================================
	// |                                            |
	// |                    Tools                   |
	// |                                            |
	// ==============================================
	
	protected void restoreBoundHandle(Constraint cst, IntVar var, int newVal, int bound, JoinPoint.StaticPart jp) {
		elementOpening("<restore " + eventAttributes(jp) + ">");
		elementOpening("<delta vident=\"v" + var.varNumber + "\">");
		if (bound == 0)
			singleElementLn("<range from=\"" + newVal + "\" to=\"" + (var.getInf()-1) + "\"/>");
		else
			singleElementLn("<range from=\"" + (var.getSup()+1) + "\" to=\"" + newVal + "\"/>");
		elementClosing("</delta>");
		singleElementLn("<update  vident=\"v" + var.varNumber + "\" types=\"" +
				(bound == 0?"min":"max") + "\"/>");
		elementClosing("</restore>");
	}
	
	protected void printDelta(IntVar var, int bound) {
		if (var.hasEnumeratedDomain()) {
			singleElement("<values>");
			if (bound == 0) {
				int i = 0;
				while (i < previousDomain.length && previousDomain[i] < var.getInf()) {
					stream.print(previousDomain[i] + " ");
					i++;
				}
			} else {
				int i = 0;
				while (i < previousDomain.length && previousDomain[i] < var.getSup()) {
					i++;
				}
				i++;
				while (i < previousDomain.length) {
					stream.print(previousDomain[i] + " ");
					i++;
				}
			}
			stream.println("</values>");
		} else {
			if (bound == 0) {
				singleElementLn("<range from=\"" + previousDomain[0] + "\" to=\"" + (var.getInf() - 1) + "\"/>");
			} else {
				singleElementLn("<range from=\"" + (var.getSup() + 1) + "\" to=\"" + previousDomain[1] + "\"/>");
			}
		}		
	}
	
	protected void updateBoundHandle(Constraint cst, IntVar var, Explanation e, int bound, JoinPoint.StaticPart jp) {
		elementOpening("<reduce " + eventAttributes(jp) + 
			(cst != null?"cident=\"c" + cst.constNumber + "\"":"") + ">");
		elementOpening("<delta>");
		printDelta(var, bound);
		elementClosing("</delta>");
		singleElementLn("<update  vident=\"v" + var.varNumber +
			"\" types=\"" + (bound == 0?"min":"max") + "\"/>");
		elementOpening("<explanation>");
		printDelta(var, bound);
		singleElement("<constraints cidents=\"");
		java.util.BitSet csts = e.getBitSet();
		for (int i = csts.nextSetBit(0); i >= 0; i = csts.nextSetBit(i + 1)) {
			stream.print("c" + ((PalmProblem)var.getProblem()).getConstraintNb(i).constNumber + " ");
		}
		stream.println("\"/>");
		elementClosing("</explanation>");
		elementClosing("</reduce>");
		stream.flush();
	}
	
	protected void removeValHandle(Constraint cst, IntVar var, Explanation e, int val, JoinPoint.StaticPart jp) { 
		elementOpening("<reduce " + eventAttributes(jp) + 
			(cst != null?"cident=\"c" + cst.constNumber + "\"":"") + ">");
		elementOpening("<delta>");
		singleElementLn("<values>"+val+"</values>");
		elementClosing("</delta>");
		singleElementLn("<update  vident=\"v" + var.varNumber +
			"\" types=\"val\"/>");
		elementOpening("<explanation>");
		singleElementLn("<values>"+val+"</values>");
		singleElement("<constraints cidents=\"");
		java.util.BitSet csts = e.getBitSet();
		for (int i = csts.nextSetBit(0); i >= 0; i = csts.nextSetBit(i + 1)) {
			stream.print("c" + ((PalmProblem)var.getProblem()).getConstraintNb(i).constNumber + " ");
		}
		stream.println("\"/>");
		elementClosing("</explanation>");
		elementClosing("</reduce>");
		stream.flush();
	}
	
	protected String solverName() {
		return "Palm";
	}
	
	protected int getCurrentDepth() {
		return 0;
	}
}
